@@ -2,6 +2,8 @@ from __future__ import annotations
 import numpy as np
 from scipy.spatial.transform import Rotation as R
 from scipy.optimize import minimize
+from math import sqrt
+from numba import njit
 from .auxiliary import html_table_from_matrix, html_table_from_vector
 from typing import Union, List, Tuple, Any, Sequence, TypeVar, Optional
 
@@ -9,17 +11,6 @@ RotationMatrixLike = Union[Sequence[Sequence[float]], np.ndarray, "RotationMatri
 VectorLike = Union[Sequence[float], np.ndarray, "Vector", "Point"]
 
 
-def normalize(vec: VectorLike) -> np.ndarray:
-    """2-norm normalize the given vector.
-    
-    Args: 
-        vec: non-normalized vector
-    Returns:
-        normalized vector
-    """
-    return np.asarray(vec) / np.linalg.norm(np.asarray(vec))
-
-
 class Frame:
     """A geometric Frame.
 
@@ -204,7 +195,7 @@ class Frame:
         Returns:
             New unit frame object
         """
-        return Frame(np.identity(3), np.zeros(3))
+        return cls(np.identity(3), np.zeros(3))
 
     @classmethod
     def from_SA_pastable_string(cls, SA_string: str) -> Frame:
@@ -221,7 +212,7 @@ class Frame:
             trans = a[:3, 3]
         except:
             raise Exception("SA string could not be read.")
-        return Frame(rot, trans)
+        return cls(rot, trans)
 
     @classmethod
     def from_extrinsic_euler_and_translations(
@@ -248,7 +239,7 @@ class Frame:
         """
         rot = R.from_euler("xyz", [theta_x, theta_y, theta_z], degrees=True).as_matrix()
         trans = [dx, dy, dz]
-        return Frame(rot, trans)
+        return cls(rot, trans)
 
     @classmethod
     def from_intrinsic_euler_and_translations(
@@ -275,7 +266,36 @@ class Frame:
         """
         rot = R.from_euler("XYZ", [theta_x, theta_y, theta_z], degrees=True).as_matrix()
         trans = [dx, dy, dz]
-        return Frame(rot, trans)
+        return cls(rot, trans)
+
+    @classmethod
+    def _from_quat_and_translations_scipy(
+        cls,
+        q0: float,
+        q1: float,
+        q2: float,
+        q3: float,
+        dx: float,
+        dy: float,
+        dz: float,
+    ) -> Frame:
+        """Frame from quaternion components (scalar last) and translations.
+
+        Args:
+            q0: quaternion component 0 (x)
+            q1: quaternion component 1 (y)
+            q2: quaternion component 2 (x)
+            q3: quaternion component 3 (scalar)
+            dx: translation along x
+            dy: translation along y
+            dz: translation along z
+
+        Returns:
+            Resulting frame
+        """
+        rot = R.from_quat([q0, q1, q2, q3]).as_matrix()
+        trans = [dx, dy, dz]
+        return cls(rot, trans)
 
     @classmethod
     def from_quat_and_translations(
@@ -302,9 +322,10 @@ class Frame:
         Returns:
             Resulting frame
         """
-        rot = R.from_quat([q0, q1, q2, q3]).as_matrix()
+        quat = normalized_quat((q0, q1, q2, q3))
+        rot = quat_as_matrix(quat)
         trans = [dx, dy, dz]
-        return Frame(rot, trans)
+        return cls(rot, trans)
 
     @classmethod
     def from_orthogonal_vectors(
@@ -324,7 +345,14 @@ class Frame:
         Returns:
             Frame: Resulting Frame
         """
-        rot = np.stack([normalize(new_x), normalize(new_y), normalize(new_z)], 1)
+        rot = np.stack(
+            [
+                normalized_vector(np.array(new_x)),
+                normalized_vector(np.array(new_x)),
+                normalized_vector(np.array(new_x)),
+            ],
+            1,
+        )
         trans = origin
         return Frame(rot, trans)
 
@@ -343,6 +371,20 @@ class Vector:
         """
         self._a: np.ndarray = np.array(v)  # storage as Numpy array
 
+    @classmethod
+    def from_array(cls, a: np.ndarray, copy=True):
+        """Initialize a vector from a numpy array of dX,dY,dZ deltas. 
+
+        Args:
+            a: A np.array containing dX,dY,dZ deltas
+        """
+        obj = cls.__new__(cls)
+        if copy:
+            obj._a = a.copy()  # storage as Numpy array
+        else:
+            obj._a = a
+        return obj
+
     def express_in_frame(
         self, new_frame: Frame, original_frame: Frame = Frame.create_unit_frame()
     ) -> Vector:
@@ -359,9 +401,11 @@ class Vector:
         """
         return express_vector_in_frame(self._a, new_frame, original_frame)
 
+    def __repr__(self) -> str:
+        return self.__str__()
+
     def __str__(self) -> str:
-        # basic string representation
-        return "<%s instance at %s> %s" % (self.__class__.__name__, id(self), self._a)
+        return "<%s at %s> %s" % ("Vector", id(self), self._a)
 
     def _repr_html_(self):
         html = html_table_from_vector(self._a, indices=["x", "y", "z"])
@@ -400,7 +444,7 @@ class Vector:
         Returns:
             Normalized vector
         """
-        return Vector(normalize(self._a))
+        return Vector.from_array(normalized_vector(self._a), copy=False)
 
     def length(self) -> float:
         """Length of the vector
@@ -408,7 +452,7 @@ class Vector:
         Returns:
             The 2-norm length of the vector.
         """
-        return np.linalg.norm(self._a)
+        return norm_L2(self._a)
 
     def transform(self, transformation: Frame) -> Vector:
         """Transform this vector by a given transformation frame.
@@ -421,9 +465,12 @@ class Vector:
         Returns:
             vector expressed in the original frame, but transformed.
         """
-        
-        #return rotate_vector(self._a, transformation._rot)
-        return Vector(transformation._rot @ self._a)
+
+        # return rotate_vector(self._a, transformation._rot)
+        # return Vector.from_array(transformation._rot @ self._a, copy=False)
+        return Vector.from_array(
+            transform_vector(transformation._rot, self._a), copy=False
+        )
 
     def __matmul__(self, other: Union[VectorLike, RotationMatrixLike]) -> float:
         return np.dot(self._a, np.asarray(other))
@@ -441,6 +488,7 @@ class Vector:
             return False
         return np.allclose(self._a, o._a, rtol=1e-10)
 
+
 class Point:
     """A Point is a container for one set of X,Y,Z coordinates.
 
@@ -455,6 +503,20 @@ class Point:
         """
         self._a: np.ndarray = np.array(p)  # storage as Numpy array
 
+    @classmethod
+    def from_array(cls, a: np.ndarray, copy=True):
+        """Initialize a Point from a numpy array of of X,Y,Z coordinate. 
+
+        Args:
+            a: A np.array containing of X,Y,Z coordinate
+        """
+        obj = cls.__new__(cls)
+        if copy:
+            obj._a = a.copy()  # storage as Numpy array
+        else:
+            obj._a = a
+        return obj
+
     def express_in_frame(
         self, new_frame, original_frame: Frame = Frame.create_unit_frame()
     ) -> Point:
@@ -471,9 +533,11 @@ class Point:
         """
         return express_point_in_frame(self._a, new_frame, original_frame)
 
+    def __repr__(self) -> str:
+        return self.__str__()
+
     def __str__(self) -> str:
-        # basic string representation
-        return "<%s instance at %s> %s" % (self.__class__.__name__, id(self), self._a)
+        return "<%s at %s> %s" % ("Point", id(self), self._a)
 
     def _repr_html_(self):
         html = html_table_from_vector(self._a, indices=["x", "y", "z"])
@@ -530,8 +594,14 @@ class Point:
         Returns:
             Point expressed in the original frame but transformed.
         """
-        #return Point(transform_points(self, transformation))
-        return Point(transformation._rot@self._a + transformation._trans)
+        # return Point(transform_points(self, transformation))
+        # return Point.from_array(
+        #     transformation._rot @ self._a + transformation._trans, copy=False
+        # )
+        return Point.from_array(
+            transform_point(transformation._rot, transformation._trans, self._a),
+            copy=False,
+        )
 
 
 class RotationMatrix:
@@ -638,8 +708,8 @@ def frame_wizard(
     assert (
         secondary_axis != primary_axis
     ), "secondary axis must not equal primary axis: choose from x,y,z"
-    primary_vec = normalize(primary_vec)
-    secondary_vec = normalize(secondary_vec)
+    primary_vec = normalized_vector(np.array(primary_vec))
+    secondary_vec = normalized_vector(np.array(secondary_vec))
     rot = np.zeros((3, 3))
     column_dict = {"x": 0, "y": 1, "z": 2}
     primary_index = column_dict.pop(primary_axis)
@@ -655,7 +725,7 @@ def frame_wizard(
     rot[:, primary_index] = primary_vec
 
     # construct tertiary and secondary axis
-    rot[:, tertiary_index] = signature_perm * normalize(
+    rot[:, tertiary_index] = signature_perm * normalized_vector(
         np.cross(primary_vec, secondary_vec)
     )
     rot[:, secondary_index] = -signature_perm * np.cross(
@@ -709,9 +779,7 @@ def express_frame_in_frame(input_frame: Frame, reference_frame: Frame):
 
 
 def express_point_in_frame(
-    point: VectorLike,
-    new_frame: Frame,
-    original_frame: Frame = Frame.create_unit_frame(),
+    point: VectorLike, new_frame: Frame, original_frame: Optional[Frame] = None,
 ) -> Point:
     """Express a point in a different frame.
 
@@ -725,7 +793,10 @@ def express_point_in_frame(
     Returns:
         Point expressed in `new_frame`.
     """
-    trafo = transformation_between_frames(original_frame, new_frame)
+    if original_frame is None:
+        trafo = new_frame
+    else:
+        trafo = transformation_between_frames(original_frame, new_frame)
     return Point(
         (np.asarray(point) - trafo._trans) @ trafo._rot
     )  # multiplication to the right is the same as with transpose to the left
@@ -734,7 +805,7 @@ def express_point_in_frame(
 def express_points_in_frame(
     points: Sequence[VectorLike],
     new_frame: Frame,
-    original_frame: Frame = Frame.create_unit_frame(),
+    original_frame: Optional[Frame] = None,
 ) -> Sequence[VectorLike]:
     """Express points in a different frame.
 
@@ -748,16 +819,17 @@ def express_points_in_frame(
     Returns:
         Points expressed in `new_frame`.
     """
-    trafo = transformation_between_frames(original_frame, new_frame)
+    if original_frame is None:
+        trafo = new_frame
+    else:
+        trafo = transformation_between_frames(original_frame, new_frame)
     return (
         np.asarray(points) - trafo._trans
     ) @ trafo._rot  # multiplication to the right is the same as with transpose to the left
 
 
 def express_vector_in_frame(
-    vector: VectorLike,
-    new_frame: Frame,
-    original_frame: Frame = Frame.create_unit_frame(),
+    vector: VectorLike, new_frame: Frame, original_frame: Optional[Frame] = None,
 ) -> Vector:
     """Express a vector in a different frame.
 
@@ -766,12 +838,15 @@ def express_vector_in_frame(
     Args:
         vector: 3x1 vector object
         new_frame: Frame to express this vector in. 
-        original_frame: Reference frame where the vector is specified in.
+        original_frame: Reference frame where the vector is specified in (defaults to UnitFrame).
 
     Returns:
         Vector expressed in `new_frame`.
     """
-    trafo = transformation_between_frames(original_frame, new_frame)
+    if original_frame is None:
+        trafo = new_frame
+    else:
+        trafo = transformation_between_frames(original_frame, new_frame)
     return Vector(np.asarray(vector) @ trafo._rot)
 
 
@@ -788,6 +863,16 @@ def rotate_vector(vec: VectorLike, rot: RotationMatrixLike) -> Vector:
     return Vector(np.asarray(rot) @ np.asarray(vec))
 
 
+@njit
+def transform_vector(rot, vec):
+    return cast_vec_to_array(mult_mat_vec(rot, vec))
+
+
+@njit
+def transform_point(rot, trans, p):
+    return cast_vec_to_array(add_vec_vec(mult_mat_vec(rot, p), trans))
+
+
 def transform_points(
     points: Union[VectorLike, Sequence[VectorLike]], trafo: Frame
 ) -> np.ndarray:
@@ -796,7 +881,7 @@ def transform_points(
 
 
 def distance_between_points(pointA: VectorLike, pointB: VectorLike) -> float:
-    return np.linalg.norm(np.asarray(pointA) - np.asarray(pointB))
+    return norm_L2(np.asarray(pointA) - np.asarray(pointB))
 
 
 def minimize_points_to_points_distance(
@@ -841,3 +926,136 @@ def minimize_points_to_points_distance(
         return t, m
     else:
         return t
+
+
+@njit
+def normalized_quat(q) -> Tuple[float, float, float, float]:
+    """ Return unit quaternion
+    
+    by dividing by Euclidean (L2) norm
+    
+    Args:
+        q array with elements
+            q0: quaternion component 0 (x)
+            q1: quaternion component 1 (y)
+            q2: quaternion component 2 (x)
+            q3: quaternion component 3 (scalar)
+    
+    Returns: 
+        array shape (4,) vector divided by L2 norm
+    """
+    n = norm_L2(q)
+    return (q[0] / n, q[1] / n, q[2] / n, q[3] / n)
+
+
+@njit
+def normalized_vector(vec) -> np.ndarray:
+    """ Return unit vector
+    
+    by dividing by Euclidean (L2) norm
+    
+    Args:
+        vec array with elements x,y,z
+    
+    Returns: 
+        array shape (3,) vector divided by L2 norm
+    """
+    res = np.empty(3)
+    n = norm_L2(vec)
+    res[0] = vec[0] / n
+    res[1] = vec[1] / n
+    res[2] = vec[2] / n
+    return res
+
+
+@njit
+def norm_L2(vec):
+    s = 0
+    for v in vec:
+        s += v ** 2
+    return sqrt(s)
+
+
+@njit
+def quat_as_matrix(unit_quat):
+    """Represent unit quaternion as rotation matrix.
+    
+    Method from scipy.spatial.transform.Rotation, 
+    jit-compiled by numba for speedup. 
+
+    Returns
+    -------
+    matrix : ndarray, shape (3, 3) or (N, 3, 3)
+        Shape depends on shape of inputs used for initialization.
+    References
+    ----------
+    .. [1] https://en.wikipedia.org/wiki/Rotation_matrix#In_three_dimensions
+    Examples
+    --------
+    >>> from scipy.spatial.transform import Rotation as R
+    Represent a single rotation:
+    >>> r = R.from_rotvec([0, 0, np.pi/2])
+    >>> r.as_matrix()
+    array([[ 2.22044605e-16, -1.00000000e+00,  0.00000000e+00],
+           [ 1.00000000e+00,  2.22044605e-16,  0.00000000e+00],
+           [ 0.00000000e+00,  0.00000000e+00,  1.00000000e+00]])
+    """
+    x = unit_quat[0]
+    y = unit_quat[1]
+    z = unit_quat[2]
+    w = unit_quat[3]
+
+    x2 = x * x
+    y2 = y * y
+    z2 = z * z
+    w2 = w * w
+
+    xy = x * y
+    zw = z * w
+    xz = x * z
+    yw = y * w
+    yz = y * z
+    xw = x * w
+
+    matrix = np.empty((3, 3))
+
+    matrix[0, 0] = x2 - y2 - z2 + w2
+    matrix[1, 0] = 2 * (xy + zw)
+    matrix[2, 0] = 2 * (xz - yw)
+    matrix[0, 1] = 2 * (xy - zw)
+    matrix[1, 1] = -x2 + y2 - z2 + w2
+    matrix[2, 1] = 2 * (yz + xw)
+    matrix[0, 2] = 2 * (xz + yw)
+    matrix[1, 2] = 2 * (yz - xw)
+    matrix[2, 2] = -x2 - y2 + z2 + w2
+
+    return matrix
+
+
+@njit
+def add_vec_vec(v1, v2):
+    return (v1[0] + v2[0], v1[1] + v2[1], v1[2] + v2[2])
+
+
+@njit
+def mult_vec_sca(v, s):
+    return (v[0] * s, v[1] * s, v[2] * s)
+
+
+@njit
+def dot_vec_vec(v1, v2):
+    return v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2]
+
+
+@njit
+def mult_mat_vec(m, v):
+    return (dot_vec_vec(m[0], v), dot_vec_vec(m[1], v), dot_vec_vec(m[2], v))
+
+
+@njit
+def cast_vec_to_array(vec):
+    a = np.empty(3)
+    a[0] = vec[0]
+    a[1] = vec[1]
+    a[2] = vec[2]
+    return a
